from pylipd.lipd_to_rdf import LipdToRDF
import multiprocessing as mp

def convert_to_rdf(lipdfile, rdffile, collection_id=None):
    converter = LipdToRDF(collection_id)    
    """Worker that converts one lipdfile to an rdffile"""
    try:
        converter.convert(lipdfile, rdffile)
    except Exception as e:
        print(f"ERROR: Could not convert LiPD file {lipdfile} to RDF")            
        raise e

def multi_convert_to_rdf(filemap, collection_id=None):
    """Create a pool to convert all lipdfiles to rdffiles"""
    pool = mp.Pool(mp.cpu_count())
    args = [(lipdfile, rdffile, collection_id) for lipdfile, rdffile in filemap.items()]
    pool.starmap(convert_to_rdf, args, chunksize=1)
    pool.close()


def convert_to_pickle(lipdfile, tofile, collection_id=None):
    converter = LipdToRDF(collection_id)    
    """Worker that converts one lipdfile to an rdffile"""
    try:
        converter.convert(lipdfile, tofile, type="pickle")
    except Exception as e:
        print(f"ERROR: Could not convert LiPD file {lipdfile} to RDF")            
        raise e

def multi_convert_to_pickle(filemap, collection_id=None):
    """Create a pool to convert all lipdfiles to picklefiles"""
    pool = mp.Pool(mp.cpu_count())
    args = [(lipdfile, tofile, collection_id) for lipdfile, tofile in filemap.items()]
    pool.starmap(convert_to_pickle, args, chunksize=1)
    pool.close()



from pylipd.lipd import LiPD

####################
# TODO:
# - Edit local LiPD & update endpoint
####################

if __name__=="__main__":
    local_lipd_dir = "/Users/varun/git/LiPD/PyLiPD/data/lpd"
    remote_lipd_endpoint = "https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2"

    '''
    # Convert LiPD files to RDF    
    lipd.convert_lipd_dir_to_rdf(
        local_lipd_dir,
        local_lipd_dir+".nq")
    
    # Convert one LiPD file to RDF
    convert_to_rdf(
        "/Users/varun/git/LiPD/PyLiPD/data/lpd/MD98_2181.Stott.2007.lpd",
        "/Users/varun/git/LiPD/PyLiPD/data/MD98_2181.Stott.2007.nq"
    )
    '''

    lipd = LiPD()
    data = ['https://lipdverse.org/data/TjhHrDv0LQ4aazHolZkR/1_0_0//Ocn-WEqPacific.Stott.2007.lpd']
    lipd.load(data)

    ts_list = lipd.get_timeseries(lipd.get_all_dataset_names())
    for dsname, tsos in ts_list.items():
        for tso in tsos:
            if 'paleoData_variableName' in tso:
                print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])    

    # Load Datasets (from Local and Remote)
    dsnames = ["MD98_2181.Stott.2007"]
    remote_dsnames = ["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001"]

    # Load from local
    lipd = LiPD()
    lipdfiles = [local_lipd_dir + "/" + dsname + ".lpd" for dsname in dsnames]
    #print(lipdfiles)
    
    lipd.load(lipdfiles)
    #lipd.load_from_dir(local_lipd_dir)
    print(lipd.get_all_dataset_names())

    #lipd.load(["/Users/varun/Downloads/Arc-LakeNataujärvi.Ojala.2005.lpd"])
    #print(lipd.get_all_dataset_names())
    
    for dsname in lipd.get_all_dataset_names():
        json = lipd.get_lipd(dsname)
        print(json['pub'])

    ts_list = lipd.get_timeseries(lipd.get_all_dataset_names())
    for dsname, tsos in ts_list.items():
        for tso in tsos:
            if 'paleoData_variableName' in tso:
                print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])

    # Fetch LiPD data from remote RDF Graph
    lipd.set_endpoint(remote_lipd_endpoint)
    lipd.load_remote_datasets(remote_dsnames)

    # Convert to TSO object (as before)
    ts_list_remote = lipd.get_timeseries(lipd.get_all_dataset_names())
    for dsname, tsos in ts_list_remote.items():
        for tso in tsos:
            print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])

    print(lipd.get_all_dataset_names())
    poplipd = lipd.pop(remote_dsnames[0])
    print("After popping..")
    print(lipd.get_all_dataset_names())
    print("Popped..")
    print(poplipd.get_all_dataset_names())
    
    '''
    print(lipd.get_all_dataset_names())    
    datasets = lipd.get_datasets(dsnames=dsnames)
    print("Fetched..")
    for ds in datasets:
        print(ds['id'])
        #print(json.dumps(ds, indent=3))
    '''

    # Usage
    # - Just look at https://pyleoclim-util.readthedocs.io/en/master/core/api.html#lipdseries-pyleoclim-lipdseries
    # - Implementing Series, MultipleSeries, EnsemebleSeries, Lipd, LipdSeries
    # - Pyleoclim
    #     - Given: Variable name, optional Dataset IDs
    #       - Return time series data for the  matching variables (optional fuzzy match)
    #     - Given: Proxy, optional Dataset IDs
    #       - Return time series data for the  variables with matching proxies (optional fuzzy match)

    #   - Time Series Data consists of:
    #       - Variable value, Variable name, Variable unit, Time axis name (age/year), Time axis unit, Time axis values, Lat, Long, Dataset ID

    #   - Lat, Long, Elevation
    #   - Get the ensemble table given a dataset ID & age Year
    # - Own metadata


"""
The LiPD class describes a `LiPD (Linked Paleo Data) <https://cp.copernicus.org/articles/12/1093/2016/cp-12-1093-2016.html>`_ object. It contains an `RDF <https://www.w3.org/RDF/>`_ Graph which is serialization of the LiPD data into an RDF graph containing terms from the `LiPD Ontology <http://linked.earth/Ontology/release/core/1.2.0/index-en.html>`
How to browse and query LiPD objects is described in a short example below, while `this notebook <https://nbviewer.jupyter.org/github/LinkedEarth/pylipd/blob/master/example_notebooks/pylipd_tutorial.ipynb>`_ demonstrates how to use PyLiPD to view and query LiPD datasets.
"""

import os
import pickle
import re
import copy
import os.path
import tempfile
import pandas as pd

from rdflib import ConjunctiveGraph, Namespace
from pylipd.multi_processing import multi_convert_to_pickle, multi_convert_to_rdf

from pylipd.rdf_to_lipd import RDFToLiPD
from pylipd.legacy_utils import LiPD_Legacy
from pylipd.utils import sanitizeId, sparql_results_to_df

from .globals.urls import NSURL, ONTONS


class LiPD:
    '''The LiPD class describes a `LiPD (Linked Paleo Data) <https://cp.copernicus.org/articles/12/1093/2016/cp-12-1093-2016.html>`_ object. It contains an `RDF <https://www.w3.org/RDF/>`_ Graph which is serialization of the LiPD data into an RDF graph containing terms from the `LiPD Ontology <http://linked.earth/Ontology/release/core/1.2.0/index-en.html>`
    How to browse and query LiPD objects is described in a short example below.

    Examples
    --------
    In this example, we read an online LiPD file and convert it into a time series object dictionary.

    .. ipython:: python
        :okwarning:
        :okexcept:

        from pylipd.lipd import LiPD

        lipd = LiPD()        
        lipd.load(["https://lipdverse.org/data/LCf20b99dfe8d78840ca60dfb1f832b9ec/1_0_1//Nunalleq.Ledger.2018.lpd"])
        
        ts_list = lipd.get_timeseries(lipd.get_all_dataset_names())

        for dsname, tsos in ts_list.items():
            for tso in tsos:
                if 'paleoData_variableName' in tso:
                    print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])
    '''
    def __init__(self, graph=None):
        if graph is None:
            self._initialize_graph()
        else:
            self.graph = graph

    def _initialize_graph(self):
        self.graph = ConjunctiveGraph()
        self.graph.bind("le", Namespace(ONTONS))        
        #self.graph.bind("", Namespace(NS))

    def load_from_dir(self, dir_path, collection_id=None):
        '''Load LiPD files from a directory
        Note: This function creates multiple process to process lipd files in parallel, therefore it is important that this call be made under the "__main__" process

        Parameters
        ----------

        dir_path : str
            path to the directory containing lipd files

        collection_id : str
            (Optional) set a collection id for all lipd files in the directory

        Examples
        --------
        In this example, we load LiPD files from a directory.

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                lipd = LiPD()        
                lipd.load_from_dir("../examples/data")

                print(lipd.get_all_dataset_names())
        '''
        if not os.path.isdir(dir_path):
            print(f"Directory {dir_path} does not exist")
            return

        lipdfiles = []
        for path in os.listdir(dir_path):
            file_path = os.path.join(dir_path, path)
            if os.path.isfile(file_path) and path.endswith(".lpd"):
                lipdfiles.append(file_path)
        self.load(lipdfiles, collection_id)


    # Allows loading http locations
    def load(self, lipdfiles, collection_id=None):
        '''Load LiPD files. 
        Note: This function creates multiple process to process lipd files in parallel, therefore it is important that this call be made under the "__main__" process

        Parameters
        ----------

        lipdfiles : array
            array of paths to lipd files (the paths could also be urls)

        collection_id : str
            (Optional) set a collection id for all lipd files in the directory

        Examples
        --------
        In this example, we load LiPD files for an array of paths.

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                lipd = LiPD() 
                lipd.load([
                    "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                    "../examples/data/MD98_2181.Stott.2007.lpd",
                    "../examples/data/Ant-WAIS-Divide.Severinghaus.2012.lpd",
                    "https://lipdverse.org/data/LCf20b99dfe8d78840ca60dfb1f832b9ec/1_0_1/Nunalleq.Ledger.2018.lpd"                    
                ])            

                print(lipd.get_all_dataset_names())
        '''        
        if type(lipdfiles) is not list:
            lipdfiles = [lipdfiles]
            
        filemap = {}
        for lipdfile in lipdfiles:
            if not os.path.isfile(lipdfile) and not lipdfile.startswith("http"):
                print(f"File {lipdfile} does not exist")
                continue

            picklefile = tempfile.NamedTemporaryFile().name
            filemap[lipdfile] = picklefile
        
        print(f"Loading {len(filemap.keys())} LiPD files" + (" from Collection: {collection_id}" if collection_id else ""))
        
        multi_convert_to_pickle(filemap, collection_id)
        print("Conversion to RDF done..")

        print("Loading RDF into graph")
        for lipdfile in lipdfiles:
            picklefile = filemap[lipdfile]
            if os.path.exists(picklefile):
                with open(picklefile, 'rb') as f:
                    subgraph = pickle.load(f)
                    self.graph.addN(subgraph.quads())
                os.remove(picklefile)
        print("Loaded..")


    def clear(self):
        '''Clears the graph'''
        self._initialize_graph()


    def set_endpoint(self, endpoint):
        '''Sets a SparQL endpoint for a remote Knowledge Base (example: GraphDB)

        Parameters
        ----------

        endpoint : str
            URL for the SparQL endpoint 

        Examples
        --------
        
        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd_remote = LiPD()
                lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2")
                lipd_remote.load_remote_datasets(["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001", "MD98_2181.Stott.2007", "Ant-WAIS-Divide.Severinghaus.2012"])
                print(lipd_remote.get_all_dataset_names())

        '''
        self.endpoint = endpoint


    def convert_lipd_dir_to_rdf(self, lipd_dir, rdf_file, collection_id=None):
        '''Convert a directory containing LiPD files into a single RDF file (to be used for uploading to Knowledge Bases like GraphDB)

        Parameters
        ----------

        lipd_dir : str
            Path to the directory containing lipd files

        rdf_file : str
            Path to the output rdf file

        collection_id : str
            (Optional) set a collection id for all lipd files in the directory

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd = LiPD()

                lipd.convert_lipd_dir_to_rdf("../examples/data", "all-lipd.nq")
        '''

        filemap = {}
        for path in os.listdir(lipd_dir):
            fullpath = os.path.join(lipd_dir, path)
            tmp_rdf_file = tempfile.NamedTemporaryFile().name
            filemap[fullpath] = tmp_rdf_file
        
        print(f"Starting conversion of {len(filemap.keys())} LiPD files")

        multi_convert_to_rdf(filemap, collection_id)
        
        print("Conversion to RDF done..")

        print("Writing to main RDF file..")
        with open(rdf_file, "a") as fout:
            for lipdfile in filemap.keys():
                tmp_rdf_file = filemap[lipdfile]
                if os.path.exists(tmp_rdf_file):
                    fin = open(tmp_rdf_file, "r")
                    data = fin.read();
                    fin.close()
                    fout.write(data)
                    os.remove(tmp_rdf_file)
            fout.close()
        print("Written..")


    def query(self, query, remote=False, result="sparql"):
        '''Once LiPD files or loaded into the graph (or remote endpoint set), one can make SparQL queries to the graph

        Parameters
        ----------

        query : str
            SparQL query

        remote: bool
            (Optional) If set to True, the query will be made to the remote endpoint (if set)

        result : str
            (Optional) Result return type

        Returns
        -------

        result : dict
            Dictionary of sparql variable and binding values
        
        result_df : pandas.Dataframe
            Return the dictionary above as a pandas.Dataframe
    
        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                lipd = LiPD()
                lipd.load([
                    "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                    "../examples/data/MD98_2181.Stott.2007.lpd"
                ])
                query = """PREFIX le: <http://linked.earth/ontology#>
                        select (count(distinct ?ds) as ?count) where { 
                            ?ds a le:Dataset .
                            ?ds le:hasUrl ?url
                        }"""
                result, result_df = lipd.query(query)
                result_df
        '''

        if remote and self.endpoint:
            matches = re.match(r"\s*SELECT\s+(.+)\s+WHERE\s+{(.+)}\s*", query, re.DOTALL)
            if matches:
                vars = matches.group(1)
                where = matches.group(2)
                query = f"SELECT {vars} WHERE {{ SERVICE <{self.endpoint}> {{ {where} }} }}"   
        
        result = self.graph.query(query)
        result_df = sparql_results_to_df(result)
        
        return result, result_df 

    # def whatArchive(self):
        
    #     archive_query = """PREFIX le: <http://linked.earth/ontology#>
    #             select ?archiveType (count(distinct ?archiveType) as ?count) where { 
    #                 ?ds a le:Dataset .
    #                 ?ds le:hasUrl ?url
    #             }"""

    def load_remote_datasets(self, dsnames):
        '''Loads remote datasets into cache if a remote endpoint is set

        Parameters
        ----------

        dsnames : array
            array of dataset names

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd_remote = LiPD()
                lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2")
                lipd_remote.load_remote_datasets(["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001", "MD98_2181.Stott.2007", "Ant-WAIS-Divide.Severinghaus.2012"])
                print(lipd_remote.get_all_dataset_names())
        '''
        if not self.endpoint:
            raise Exception("No remote endpoint")
        
        if type(dsnames) is not list:
            dsnames = [dsnames]
            
        if dsnames == None or len(dsnames) == 0:
            raise Exception("No dataset names to cache")
        dsnamestr = (' '.join('<' + NSURL + "/" + dsname + '>' for dsname in dsnames))
        print("Caching datasets from remote endpoint..")
        qres, qres_df = self.query(f"SELECT ?s ?p ?o ?g WHERE {{ GRAPH ?g {{ ?s ?p ?o }} VALUES ?g {{ {dsnamestr} }} }}", remote=True)

        # Reinitialize graph
        # self._initialize_graph()
        for row in qres:
            self.graph.add((row.s, row.p, row.o, row.g))
        print("Done..")

    def get_timeseries(self, dsnames):
        '''Get Legacy LiPD like Time Series Object (tso)

        Parameters
        ----------

        dsnames : array
            array of dataset id or name strings

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd_remote = LiPD()
                lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2")
                ts_list = lipd_remote.get_timeseries(["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001", "MD98_2181.Stott.2007", "Ant-WAIS-Divide.Severinghaus.2012"])
                for dsname, tsos in ts_list.items():
                    for tso in tsos:
                        if 'paleoData_variableName' in tso:
                            print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])
        '''
        ts = self._get_timeseries(dsnames)
        return ts

    def _get_timeseries(self, dsnames):
        timeseries = {}
        for dsname in dsnames:
            converter = RDFToLiPD()
            d = converter.convert(dsname, self.graph)
            print("Extracting timeseries from dataset: " + dsname + " ...")
            if len(d.items()):
                tss = LiPD_Legacy().extract(d)
                timeseries[dsname] = tss
        return timeseries

    def get_lipd(self, dsname):
        '''Get LiPD json for a dataset

        Parameters
        ----------

        dsname : str
            dataset id

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd_remote = LiPD()
                lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2")
                dsname = "Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001"
                lipd_remote.load_remote_datasets([dsname])
                lipd_json = lipd_remote.get_lipd(dsname)
                print(lipd_json)
        '''           
        converter = RDFToLiPD()            
        return converter.convert(dsname, self.graph)

    def pop(self, dsname, collection_id=None):
        '''Removes a dataset from the graph and returns a LiPD object

        Parameters
        ----------

        dsname : str
            Path to the directory containing lipd files

        collection_id : str
            (Optional) collection id for the dataset

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd = LiPD()
                lipd.load([
                    "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                    "../examples/data/MD98_2181.Stott.2007.lpd"
                ])
                all_datasets = lipd.get_all_dataset_names()
                print("Loaded datasets: " + str(all_datasets))
                popped = lipd.pop(all_datasets[0])
                print("Loaded datasets after pop: " + str(lipd.get_all_dataset_names()))
                print("Popped dataset: " + str(popped.get_all_dataset_names())        
        '''        
        graphurl = NSURL + "/" + dsname
        if collection_id:
            graphurl = NSURL + "/" + collection_id + "/" + dsname
        subgraph = copy.deepcopy(self.graph.get_context(graphurl))
        self.graph.remove((None, None, None, graphurl))
        return LiPD(subgraph)

    def remove(self, dsname, collection_id=None):
        '''Removes a dataset from the graph

        Parameters
        ----------

        dsname : str
            Path to the directory containing lipd files

        collection_id : str
            (Optional) collection id for the dataset

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd = LiPD()
                lipd.load([
                    "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                    "../examples/data/MD98_2181.Stott.2007.lpd"
                ])
                all_datasets = lipd.get_all_dataset_names()
                print("Loaded datasets: " + str(all_datasets))
                lipd.remove(all_datasets[0])
                print("Loaded datasets after remove: " + str(lipd.get_all_dataset_names()))
        '''
        graphurl = NSURL + "/" + dsname
        if collection_id:
            graphurl = NSURL + "/" + collection_id + "/" + dsname
        self.graph.remove((None, None, None, graphurl))       

    def get_rdf(self):
        '''Returns RDF serialization of the current Graph
        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd = LiPD()
                lipd.load([
                    "../examples/data/MD98_2181.Stott.2007.lpd"
                ])
                print(lipd.get_rdf())
        '''
        
        return self.graph.serialize(format='nquads')


    def get_all_dataset_names(self):
        '''Get all Dataset Names
        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            from pylipd.lipd import LiPD

            if __name__=="__main__":
                # Fetch LiPD data from remote RDF Graph
                lipd = LiPD()
                lipd.load([
                    "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                    "../examples/data/MD98_2181.Stott.2007.lpd"
                ])
                print(lipd.get_all_dataset_names())
        '''        
        query = f"""
            SELECT ?dsname WHERE {{ 
                ?ds a le:Dataset .
                ?ds le:name ?dsname
            }}
            """
        qres, qres_df = self.query(query)
        return [sanitizeId(row.dsname) for row in qres]

    def search_datasets(variableName=[ ], archiveType=[ ], proxyObsType=[ ], infVarType = [ ], sensorGenus=[ ],
                    sensorSpecies=[ ], interpName =[ ], interpDetail =[ ], ageUnits = [ ],
                    ageBound = [ ], ageBoundType = [ ], recordLength = [ ], resolution = [ ],
                    lat = [ ], lon = [ ], alt = [ ], print_response = True, download_lipd = True,
                    download_folder = 'default'):
        pass


    def find_ensemble_table_for_variable(self, ensemble_table):
        pass

    def export_biblio(self):
        pass


"""
The RDFToLiPD class helps in converting an RDF Graph to a LiPD file.
It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion
"""

import copy
import re
import json

from rdflib.graph import ConjunctiveGraph, URIRef

from .globals.urls import NSURL
from .globals.blacklist import REVERSE_BLACKLIST
from .globals.schema import SCHEMA

from .utils import ucfirst, lcfirst, unCamelCase, unzip_string

class RDFToLiPD:
    """
    The RDFToLiPD class helps in converting an RDF Graph to a LiPD file.
    It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion

    Parameters
    ----------
    collection_id : str
        (Optional) set a collection id for the lipd file    
    """    
    def __init__(self, collection_id=None):
        self.graph = ConjunctiveGraph()
        self.lipd_csvs = {}
        self.collection_id = collection_id
        self.graphurl = NSURL
        self.namespace = NSURL + "#"
        if self.collection_id:
            self.namespace = NSURL + "/" + collection_id + "#"

    def convert(self, id, graph):
        '''Convert RDF graph to LiPD json

        Parameters
        ----------

        graph : rdflib.ConjunctiveGraph
            the RDF graph object

        '''
        self.graph = graph
        self.schema = copy.deepcopy(SCHEMA)
        self.rschema = self._get_schema_reverse_map()

        self.allfacts = {}
        self._get_indexed_facts(self.namespace + id)
        
        lipd = self._convert_to_lipd(self.namespace + id, "Dataset", "Dataset", pagesdone=[])
        lipd = self._post_processing(lipd)
        return lipd

    def _get_property_details(self, pname, schema) :
        details = { "name": pname }
        # Create property
        if schema and (pname in schema) :
            for skey,svalue in schema[pname].items() :
                details[skey] = svalue
        return details

    def _get_rdf_property_details(self, pname, fullkey, schema) :
        key = pname
        pname = lcfirst(pname)
        details = { "name": pname }
		# Check for full key in schema
        if schema and (fullkey in schema):
            for ind, svalue in schema[fullkey].items():
                details[ind] = svalue
		
        # Or check for just pname in schema
        elif schema and (key in schema):
            for ind, svalue in schema[key].items():
                details[ind] = svalue

        return details;


    def _get_schema_reverse_map(self) :
        newschema = {}
        for schid,sch in self.schema.items() :
            newsch = {}
            for prop,details in sch.items() :
                if (prop[0] == "@") :
                    continue
                
                if (("hack" in details)) :
                    continue
                
                pdetails = self._get_property_details(prop, sch)
                pname = pdetails["name"]
                pdetails["name"] = prop
                newsch[pname] = pdetails
                if (("category" in pdetails)) :
                    catpname = pname + "." + ucfirst(pdetails["category"])
                    newsch[catpname] = pdetails
                
                if (("schema" in pdetails)) :
                    schpname = pname + "." + ucfirst(pdetails["schema"])
                    newsch[schpname] = pdetails

            newschema[schid] = newsch
        
        return newschema

    def _local_name(self, url):
        return str(url).split("#")[-1]

    def _get_prop_values_from_query_result_p_o(self, qres):
        result = {}
        for row in qres:
            pname = self._local_name(row.p)
            if pname not in result:
                result[pname] = []
            value = {}
            if isinstance(row.o, URIRef):
                value["@type"] = "uri"
                value["@id"] = str(row.o)
            else:
                value["@type"] = "literal"
                value["@value"] = row.o.value
                value["@datatype"] = str(row.o.datatype)
            result[pname].append(value)
        return result

    def _get_facts(self, id):
        query = f"SELECT ?p ?o WHERE {{ <{id}> ?p ?o }}"
        qres = self.graph.query(query)
        return self._get_prop_values_from_query_result_p_o(qres)

    def _get_indexed_facts(self, id):
        if id in self.allfacts:
            return

        facts = self._get_facts(id)
        self.allfacts[id] = facts

        for pname, pfacts in facts.items():
            for pfact in pfacts:
                if pfact["@type"] == "uri":
                    if pname != "type":
                        self._get_indexed_facts(pfact["@id"])


    def _convert_to_lipd(self, id, category, schemaname, pagesdone=[]) :
        if id in self.allfacts:
            facts = self.allfacts[id]

            if id in pagesdone:
                return {}
            pagesdone.append(id)
            
            schema =  self.rschema[schemaname] if (schemaname in self.rschema) else None
            if (schemaname and not category) :
                category = schemaname
            
            if "type" in facts:
                cats = facts["type"]
                for cat in cats:
                    if (cat["@type"] == "uri") :
                        category = self._local_name(cat["@id"])
                        break
            
            obj = {
                "@id":id,
                "@category":category,
                "@schema":schemaname
            }

            for pname, pfacts in facts.items() :
                if pname in REVERSE_BLACKLIST:
                    continue
                prop = pname
                prop = re.sub("/\\s/", "_", prop)
                # Get a sample value page category, and use to make a property key
                propkey = prop
                for value in pfacts: 
                    if value["@type"] == "uri" :
                        if value["@id"] in self.allfacts :
                            pfact = self.allfacts[value["@id"]]
                            if "type" in pfact:
                                valcats = pfact["type"]
                                for valcat in valcats:
                                    if (valcat["@type"] == "uri") :
                                        valcatname = self._local_name(valcat["@id"])
                                        propkey = prop + "." + valcatname
                                        break

                details = self._get_rdf_property_details(prop, propkey, schema)
                name = details["name"]
                ptype =  details["type"] if ("type" in details) else None
                cat =  details["category"] if ("category" in details) else None
                sch =  details["schema"] if ("schema" in details) else None
                if (cat and not sch) :
                    sch = cat
                
                toJson =  details["toJson"] if ("toJson" in details) else None
                multiple =  details["multiple"] if ("multiple" in details) else False

                if len(pfacts) > 0 :
                    if (multiple) :
                        obj[name] = []
                    
                    for pfact in pfacts : 
                        if pfact["@type"] == "uri" :
                            val = self._convert_to_lipd(pfact["@id"], cat, sch, pagesdone)
                        else:
                            val = pfact["@value"]
                        
                        if (toJson) :
                            fn = getattr(self, toJson)
                            val = fn(val)
                        
                        # TODO: Handle hasValues :: Convert to File ?
                        
                        # If there is already a value present
                        # - Then this need to be marked as "multiple"
                        if (not multiple) and (name in obj) and (not type(obj[name]) is list):
                            multiple = True
                            obj[name] = [obj[name]]
                        
                        if (multiple) :
                            obj[name].append(val)
                        else : 
                            obj[name] = val

            # TODO: Fix the empty case ?
            #if (empty(Settlement.array_keys([obj]))) :
            #    return preg_replace("/_/", " ", id)
            
            return obj
        else : 
            return re.sub("/_/", " ", id)


    def _post_processing(self, obj, parent=None):
        if not(type(obj) is dict):
            return obj
		
        if not("@schema" in obj):
            return obj

        for key,value in obj.items():
            if type(value) is list:
                for i in range(len(value)):
                    obj[key][i] = self._post_processing(value[i], obj)
            else:
                obj[key] = self._post_processing(value, obj)

        schemaname = obj["@schema"]
        tschema = self.schema[schemaname] if schemaname in self.schema else None
        if tschema and "@toJson" in tschema:
            for func in tschema["@toJson"]:
                fn = getattr(self, func)
                obj = fn(obj, parent)

        if "hasValues" in obj:
            obj["values"] = json.loads(obj["hasValues"])
            del obj["hasValues"]
        
        del obj["@id"]
        del obj["@schema"]
        del obj["@category"]
        if "type" in obj:
            del obj["type"] 

        return obj

    #########################################
    # Converters
    #########################################
    def _location_to_json(self, geo, parent = None) :
        geojson = {
            "geometry":
                {
                    "coordinates" : [0, 0, 0],
                },
            "properties": {}
        }
        if "coordinates" in geo :
            latlong = geo["coordinates"].split(",")
            geojson["geometry"]["coordinates"] = [ float(latlong[1]), float(latlong[0]), float(latlong[2]) if len(latlong)>2 else 0 ]
            geojson["geometry"]["type"] = "Point"
        
        if "long" in geo :
            geojson["geometry"]["coordinates"][0] = float(geo["long"])
        
        if "lat" in geo :
            geojson["geometry"]["coordinates"][1] = float(geo["lat"])
        
        if "alt" in geo and geo["alt"] != "NA":
            geojson["geometry"]["coordinates"][2] = float(geo["alt"])
        
        for prop,value in geo.items() :
            if prop[0] == "@" :
                continue
            
            if prop == "locationType" :
                geojson["type"] = geo["locationType"]
            else : 
                if prop == "coordinates" or prop == "coordinatesFor":
                    # Ignore
                    pass
                else : 
                    if re.search(r"^(geo|wgs84):", prop) :
                        # Ignore
                        pass
                    else : 
                        geojson["properties"][prop] = value
        
        return geojson


    def _get_google_spreadsheet_key(self, url:str, parent = None) :
        return url.replace("https://docs.google.com/spreadsheets/d/", "")

    def _remove_found_in_table(self, var, parent = None) :
        if (("foundInTable" in var)) :
            del var["foundInTable"]
        return var

    def __get_variable_archive_types(self, item, atypes) :
        if type(item) is dict:
            nitem = {}
            for key,value in item.items() :
                if (key == "archiveType") :
                    atypes[value] = 1
                else : 
                    [nitem[key], atypes] = self.__get_variable_archive_types(value, atypes)
            
        elif type(item) is list:
            nitem = []
            for value in item :
                [nit, atypes] = self.__get_variable_archive_types(value, atypes)
                nitem.append(nit)
                
        else : 
            nitem = item
        return [nitem, atypes]

    def _get_variable_archive_types(self, var, parent = None) :
        [var, atypes] = self.__get_variable_archive_types(var, {})
        for atype,ok in atypes.items() :
            var["archiveType"] = atype
        return var

    def _get_lipd_archive_type(self, archiveType):
        return unCamelCase(archiveType)

    def _extract_from_proxy_system(self, var, parent = None) :
        if (("hasProxySystem" in var)) :
            ps = var["hasProxySystem"]
            #var["proxy"] = ps["name"]
            if (("hasProxySensor" in ps)) :
                psensor = ps["hasProxySensor"]
                if type(psensor) is dict:
                    if (("sensorGenus" in psensor)) :
                        var["sensorGenus"] = psensor["sensorGenus"]
                    
                    if (("sensorSpecies" in psensor)) :
                        var["sensorSpecies"] = psensor["sensorSpecies"]

            del var["hasProxySystem"]
        if (("measuredOn" in var)) :
            archive = var["measuredOn"]
            if type(archive) is dict and "@category" in archive:
                archiveType = archive["@category"]
                var["archiveType"] = self._get_lipd_archive_type(archiveType)
            elif type(archive) is str:
                var["archiveType"] = self._get_lipd_archive_type(archive)
            del var["measuredOn"]
        return var

    def _unwrap_uncertainty(self, var, parent = None) :
        if (("hasUncertainty" in var)) :
            unc = var["hasUncertainty"]
            if (("hasValue" in unc)) :
                var["uncertainty"] = float(unc["hasValue"])
                del unc["hasValue"]
            
            for key,value in unc.items() :
                if (key[0] != "@") :
                    var[key] = value
                
            del var["hasUncertainty"]
        return var

    def _unwrap_integration_time(self, interp, parent = None) :
        if (("integrationTime" in interp)) :
            intime = interp["integrationTime"]
            if (("hasValue" in intime)) :
                interp["integrationTime"] = float(intime["hasValue"])
                del intime["hasValue"]
            
            for key,value in intime.items() :
                if (key[0] != "@") :
                    interp["integrationTime" + ucfirst(key)] = value

            del interp["hasIntegrationTime"]
        return interp

    def _collect_variables_by_id(self, item, arr) :
        if not (type(item) is dict):
            return arr

        # Data is a Hash
        if ("@category" in item) and ("@id" in item) and re.match("/Variable\$/", item["@category"]) :
            arr[item["@id"]] = item
        else : 
            for key,value in item.items() :
                if (key[0] != "@") :
                    arr = self._collect_variables_by_id(item[key], arr)
        return arr

    def _set_variable_type(self, var, parent = None) :
        if (var["@category"] == "MeasuredVariable") :
            var["variableType"] = "measured"
        
        if (var["@category"] == "InferredVariable") :
            var["variableType"] = "inferred"
        return var

    def _remove_depth_property(self, val, parent = None) :
        if (("takenAtDepth" in val)) :
            del val["takenAtDepth"]
        return val

    def _create_publication_identifier(self, pub, parent = None) :
        identifiers = []
        if (("hasDOI" in pub)) :
            identifier = {}
            identifier["type"] = "doi"
            identifier["id"] = pub["hasDOI"]
            if (("link" in pub)) :
                for link in pub["link"].values() : 
                    if (re.match("/dx.doi.org/", link)) :
                        identifier["url"] = link
                del pub["link"]
            del pub["hasDOI"]
            identifiers.append(identifier)

        pub["identifier"] = identifiers
        return pub

    def _change_seasonality_type(self, interp, parent = None) :
        if (("seasonality" in interp)) :
            if type(interp["seasonality"]) is list:
                newseasonality = []
                for svalue in interp["seasonality"]:
                    newseasonality.append(float(svalue))
                interp["seasonality"] = newseasonality
        return interp

    def _values_to_array(self, resolution, parent = None) :
        if (("values" in resolution)) :
            return resolution["values"].split(",")

    def _unarray_column_number(self, var, parent = None) :
        if not var:
            return var
        if ("number" in var) and (type(var["number"]) is list) and (len(var["number"]) == 1) :
            var["number"] = var["number"][0]
        if ("number" in var) and (type(var["number"]) is str) :
            var["number"] = json.loads(var["number"])
        return var

    def _extract_variable_values(self, var, parent = None) :
        if "hasValues" in var:
            values = json.loads(var["hasValues"])
            if type(values) is dict and "base64_zlib" in values:
                values = unzip_string(values["base64_zlib"])
                var["hasValues"] = values
        return var